require "uri"
require "net/http"
require "net/https"
require "json"

def is_number?
	self.to_f == self
end

class Worldpay

	@service_key = ''
	@timeout = 0
	@disable_ssl = false
	@endpoint = ''

	@errors = {}
	@raise_errors = false

	@version = '1.2.3'

	@orderTypes = ['ECOM','MOTO','RECURRING','APM']


	def initialize(service_key='', timeout=3, raise_errors=false)
		if service_key.empty?
			raise "Please set a Service Key"
		else
			@service_key = service_key
			@timeout = timeout

			@disable_ssl = false
			@endpoint = 'https://api.worldpay.com'

			@errors = {
				'ip' => 'Invalid parameters',
				'cine' => 'php_curl was not found',
				'to' => 'Request timed out',
				'nf' => 'Not found',
				'apierror' => 'API Error',
				'uanv' => 'Worldpay is currently unavailable, please try again later',
				'contact' => 'Error contacting Worldpay, please try again later',
				'ssl' => 'You must enable SSL check in production mode',
				'verify' => 'Worldpay not verifying SSL connection',
				'orderInput' => {
								'token' => 'No token found',
								'orderCode' => 'No order_code entered',
								'orderDescription' => 'No order_description found',
								'amount' => 'No amount found, or it is not a whole number',
								'currencyCode' => 'No currency_code found',
								'name' => 'No name found',
								'billingAddress' => 'No billing_address found'
				},
				'notificationPost' => 'Notification Error: Not a post',
				'notificationUnknown' => 'Notification Error: Cannot be processed',
				'capture' => {
								'ordercode' => 'No order code entered'
				},
				'refund' => {
								'ordercode' => 'No order code entered'
				},
				'json' => 'JSON could not be decoded',
				'key' => 'Please enter your service key',
				'sslerror' => 'Worldpay SSL certificate could not be validated'
			}
			@raise_errors = raise_errors
		end
	end

	def setEndpoint(endpoint)
			@endpoint = endpoint
	end

	#Disable SSL Check ~ Use only for testing!
	def disableSSLCheck(disable = false)
		@disable_ssl = disable

	end

	#Set timeout
	def setTimeout(timeout = 3)
		@timeout = timeout
	end

	#ERRORS
	def onError(error, message=false, code=nil, httpStatusCode=nil, description=nil, customCode=nil)

		if (@raise_errors)
			raise message.to_s
		else
			if (code!=nil && httpStatusCode!=nil && description!=nil && customCode!=nil)
				return {
						'message'=>message,
						'code'=>code,
						'httpStatusCode'=>httpStatusCode,
						'description'=>description,
						'customCode'=>customCode
				}
			end
			return message.to_s
		end
	end

	#ORDERS
	def checkOrderInput(order)
		errors2 = []

		if (!order.any?)
			return onError('ip')
		end
		if (order['token']==nil)
			errors2 << @errors['orderInput']['token']
		end
		if (order['orderDescription']==nil)
			errors2 << @errors['orderInput']['orderDescription']
		end
		if (order['amount']==nil)
			errors2 << @errors['orderInput']['amount']
		end
		if (order['currencyCode']==nil)
			errors2 << @errors['orderInput']['currencyCode']
		end
		if (order['name']==nil)
			errors2 << @errors['orderInput']['name']
		end
		if (order['billingAddress']==nil)
			errors2 << @errors['orderInput']['billingAddress']
		end

		if (errors2.length > 0)
			return onError('ip', errors2.join(', '))
		end
	end



	def createOrder(order={})

		self.checkOrderInput(order)

		defaults = {
			'orderType' => 'ECOM',  #Order Type: ECOM/MOTO/RECURRING/APM
			'customerIdentifiers' => nil,
			'billingAddress' => nil,
			'deliveryAddress' => nil
		}

		order = defaults.merge(order)

		if (order['amount']!="")
			order['amount'] = order['amount'].to_i >0 ? order['amount'] : -1
			fmt = "%05.2f" % order['amount']
			cost_cents = fmt.split('.').join('')
		else
			cost_cents = order['amount']
		end

		request = {
			'token' => order['token'],
			'orderDescription' => order['orderDescription'],
			'amount' => cost_cents,
			'is3DSOrder' => order['3DS'],
			'currencyCode' => order['currencyCode'],
			'siteCode' => order['siteCode'],
			'settlementCurrency' => order['settlementCurrency'],
			'name' => order['name'],
			'shopperEmailAddress' => order['shopperEmailAddress'] ? order['shopperEmailAddress'] : "",
			'orderType' => ['ECOM','MOTO','RECURRING','APM'].include?(order['orderType']) ? order['orderType'] : 'ECOM',
			'authorizeOnly' => (!!order['authorizeOnly'] || !!order['authoriseOnly']) || false,
			'billingAddress' => order['billingAddress'],
			'deliveryAddress' => order['deliveryAddress'],
			'customerOrderCode' => order['customerOrderCode'],
			'customerIdentifiers' => order['customerIdentifiers'].present? ? JSON.parse(order['customerIdentifiers']) : nil,
			'statementNarrative' => order['statementNarrative'],
			'orderCodePrefix' => order['orderCodePrefix'],
			'orderCodeSuffix' => order['orderCodeSuffix']
		}

		if(request['is3DSOrder'] == true)
			threeDsInfo = {
				'shopperIpAddress' => order['shopperIpAddress'],
				'shopperSessionId' => order['shopperSessionId'],
				'shopperUserAgent' => order['shopperUserAgent'],
				'shopperAcceptHeader' => order['shopperAcceptHeader']
			}
			request.merge!(threeDsInfo)
		end

		response = sendRequest('orders', request.to_json, true)

		if (response['body'] && response['body']['orderCode'])
			#success
			return response
		else
			return {'response'=> response, 'error'=> onError('apierror', response.to_s)}
		end

	end

	def createAPMOrder(order={})

		self.checkOrderInput(order)

		defaults = {
			'deliveryAddress' => nil,
			'billingAddress' => nil,
			'successUrl' => nil,
			'pendingUrl' => nil,
			'failureUrl' => nil,
			'cancelUrl' => nil
		}

		order = defaults.merge(order)

		fmt = "%05.2f" % order['amount']
		cost_cents = fmt.split('.').join('')

		request = {
			'token' => order['token'],
			'orderDescription' => order['orderDescription'],
			'amount' => cost_cents,
			'currencyCode' => order['currencyCode'],
			'name' => order['name'],
			'shopperEmailAddress' => order['shopperEmailAddress'],
			'billingAddress' => order['billingAddress'],
			'deliveryAddress' => order['deliveryAddress'],
			'customerOrderCode' => order['customerOrderCode'],
			'successUrl' => order['successUrl'],
			'pendingUrl' => order['pendingUrl'],
			'failureUrl' => order['failureUrl'],
			'cancelUrl' => order['cancelUrl'],
			'statementNarrative' => order['statementNarrative'],
			'settlementCurrency' => order['settlementCurrency'],
			'customerIdentifiers' => order['customerIdentifiers'].present? ? JSON.parse(order['customerIdentifiers']) : nil,
		}

		response = sendRequest('orders', request.to_json, true)

		if (response['body']['orderCode'])
			#success
			return response
		else
			return onError('apierror', response.to_s)
		end

	end

	def authorise3DSOrder(orderCode, responseCode, threeDsInfo)
		authorize3DSOrder(orderCode, responseCode, threeDsInfo)
	end
	def authorize3DSOrder(orderCode, responseCode, threeDsInfo)
		request = {
			'threeDSResponseCode' => responseCode,
			'shopperSessionId' => threeDsInfo['shopperSessionId'],
			'shopperAcceptHeader' => threeDsInfo['shopperAcceptHeader'],
			'shopperUserAgent' => threeDsInfo['shopperUserAgent'],
			'shopperIpAddress' => threeDsInfo['shopperIpAddress']
		}

		response = sendRequest('orders/'+orderCode, request.to_json, true, 'PUT')

		if (response['body']['orderCode'])
			#success
			return response
		else
			return onError('apierror', response.to_s)
		end
	end

	#Capture Authorized Worldpay Order
	def captureAuthorisedOrder(orderCode=false, amount=false)
		captureAuthorizedOrder(orderCode, amount)
	end
	def captureAuthorizedOrder(orderCode=false, amount=false)
		if (orderCode || orderCode.is_a?(String))
			#
		else
			return onError('ip', @errors['capture']['ordercode'])
		end

		if (amount && amount.is_a?(Integer))
			json = {'captureAmount'=>amount}.to_json
		end

		sendRequest('orders/'+orderCode+'/capture', (json||false))
	end

	#Cancel Authorized Worldpay Order
	def cancelAuthorisedOrder(orderCode=false)
		cancelAuthorizedOrder(orderCode)
	end
	def cancelAuthorizedOrder(orderCode=false)
		if (orderCode || orderCode.is_a?(String))
			#
		else
			return onError('ip', @errors['capture']['ordercode'])
		end

		sendRequest('orders/'+orderCode, false, false, 'DELETE')
	end

	#Refund Worldpay order
	def refundOrder(orderCode=false, amount=false)
		if (orderCode==false || orderCode!=orderCode.to_s)
			return onError('ip', @errors['refund']['ordercode'])
		end
		if (amount!=false)
			json = {'refundAmount'=>amount}.to_json
		else
			json = false
		end
		sendRequest('orders/'+orderCode+'/refund', json)
	end

	#REQUESTS
	def sendRequest(action='', json=false, expectResponse=false, method='POST')
		uri = URI.parse(@endpoint)

		net = Net::HTTP.new(uri.host, uri.port)

		if (method=="POST")
			request = Net::HTTP::Post.new('/v1/'+action)
		elsif (method=="GET")
			request = Net::HTTP::Get.new('/v1/'+action)
		elsif (method=="PUT")
			request = Net::HTTP::Put.new('/v1/'+action)
		elsif (method=="DELETE")
			request = Net::HTTP::Delete.new('/v1/'+action)
		end

		net.use_ssl = true
		net.verify_mode = @disable_ssl ? OpenSSL::SSL::VERIFY_NONE : OpenSSL::SSL::VERIFY_PEER

		if (json != false)
			request.body = json
		end
		request.add_field("Authorization", @service_key)
		request.add_field("Content-Type", "application/json")
		request.add_field("X-wp-client-user-agent", "os.name="+RUBY_PLATFORM.to_s+";os.version="+RUBY_PLATFORM.to_s+";os.arch="+RUBY_PLATFORM.to_s+";lang.version="+RUBY_VERSION.to_s+";lib.version="+@version.to_s+";api.version=v1;lang=am;owner=worldpay")
		if (json != false)
			request.add_field("Content-Length", json.length)
		end

		net.set_debug_output $stdout #useful to see the raw messages going over the wire
		net.read_timeout = @timeout
		net.open_timeout = @timeout

		result = net.start do |http|
			http.request(request)
		end

		# raise request.body.to_s

		if (result.read_body && result.read_body.length >= 2)
			response = JSON.parse(result.read_body)
		else
			response = false
		end

		if (expectResponse && !response.present?)
			return onError('uanv', @errors['json'], 503);
		end

		if (response && response["httpStatusCode"].present?)

			if (response["httpStatusCode"] != '200')
				return onError(
					false,
					response["message"],
					result.code,
					response['httpStatusCode'],
					response['description'],
					response['customCode']
				)
			end
		elsif (expectResponse && result.code != '200')
			# If we expect a result and we have an error
			return onError('uanv', @errors['json'], 503)
		elsif (!expectResponse)
			if (result.code != '200')
				return onError('apierror', result.read_body, result.code)
			else
				response = true
			end
		end

		return {
			"code"=>result.code,
			"body"=>response
		}

	end

	#RESPONSES
	def handleResponse(response)
		response = response.to_json
	end

	#TOKENS
	#Get card details from Worldpay token
	def getStoredCardDetails(token=false)
		if (token==false || token!=token.to_s)
			return onError('ip', @errors['orderInput']['token'])
		end

		response = sendRequest('tokens/'+token, false, true, 'GET')

		if (response['body']['paymentMethod'])
			return response['body']['paymentMethod']
		else
			return false
		end
	end


	def getOrder(orderCode = false)
		if (orderCode==false || !orderCode.present?)
			return onError('ip', @errors['orderInput']['orderCode']);
		end

		response = sendRequest('orders/' + orderCode, false, true, 'GET');

		if (!response['body']['orderCode'].present?)
			return onError('apierror')
		end

		return response
	end

end
